/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2008 Funambol, Inc.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 * 
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

package com.funambol.blackberry.source;

import java.util.Stack;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.microedition.pim.PIMItem;
import javax.microedition.pim.Contact;
import javax.microedition.pim.PIMException;

import net.rim.blackberry.api.pdap.BlackBerryPIMList;
import net.rim.blackberry.api.pdap.PIMListListener;

import net.rim.device.api.system.PersistentObject;
import net.rim.device.api.system.PersistentStore;
import net.rim.device.api.system.ControlledAccessException;
import net.rim.device.api.util.Persistable;

import com.funambol.util.StaticDataHelper;

import com.funambol.blackberry.helper.PIMItemHelper;
import com.funambol.blackberry.helper.ContactHelper;
import com.funambol.syncml.spds.SourceConfig;
import com.funambol.syncml.spds.SyncException;
import com.funambol.syncml.spds.SyncItem;
import com.funambol.syncml.spds.SyncSource;

import com.funambol.syncml.protocol.SyncFilter;
import com.funambol.syncml.protocol.SyncML;
import com.funambol.syncml.protocol.SyncMLStatus;

import com.funambol.blackberry.SyncClient;
import com.funambol.util.SyncListener;
import com.funambol.util.Log;

/**
 * Base class for PIM sync source in BB. This class is a specialization of
 * BBPIMSyncSource with a workaround for a bug in RIM's APIs on OS 4.5. The
 * class can be used on any platform, but its usage is discourages unless
 * necessary.
 * The main feature provided is the ability to have different values for LUID
 * and UID. The UID is assigned by the system to any created PIM Item, the LUID
 * is the identifier of an item that is sent to the server by the Sync Engine.
 * The best choice is to have UID=LUID and this is what the standard
 * BBPIMSyncSource does. But on OS 4.5 there is a problem in the APIs that
 * sometimes are unable to clean a contact. Clearing its fields may generate
 * an unexpcted IndexOutOfBoundsException and then updates from the server
 * cannot be safely applied (in other cases there is no exception but the
 * contact is not cleared completely). To workaround this problem, this class
 * has a mapping between luid and uid and updates are implemented by removing
 * the contact and recreating it. Since the newly created contact has a
 * different UID, we need a map to know its LUID to handle operations with the
 * server.
 * At the interface level this class works with LUID. So any item that comes in
 * or goes out has a key equal to the LUID. Internally the class maps LUID to
 * UID (and viceversa) to communicate with "helpers" which operate on UIDs.
 */
public class BBPIMSyncSourceWithUpdateFix extends BBPIMSyncSource
{
    private LuidUIDMap uidMap;
    private long luidUidPersistenceStorageKey;

    public BBPIMSyncSourceWithUpdateFix(SyncClient app, SourceConfig config,
                                        BlackBerryPIMList list, PIMItemHelper helper,
                                        long persistenceStorageKey,
                                        long luidUidPersistenceStorageKey) throws SyncException
    {
        super(app, config, list, helper, persistenceStorageKey);
        this.luidUidPersistenceStorageKey = luidUidPersistenceStorageKey;
        initLuidUIDMap();
    }

    //----------------------------------------------------------- Public Methods
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#updateItem(com.funambol.syncml.spds.SyncItem)
     */
    public int updateItem(SyncItem item) throws SyncException
    {
        Log.info("[updateItem] " + item.getKey());
        totalUpdated++;
        if (app.isCancelled())  cancelSync();
        try
        {
            Log.trace("Before searching for existing PIMItem.");
            String luid = item.getKey();
            String uid = getUIDFromLuid(luid);
            PIMItem pitem = helper.findItem(uid);
            if (pitem!=null)
            {
                // We delete the item, create an empty one and add a mapping
                // Get the UID from the luid
                
                // Before removing the item we must save the fields that
                // are not synchronized so that we can restore them in the newly
                // created item
                PIMItem origItem = pitem;

                Log.debug("Deleting item");
                helper.deleteItemFromKey(uid);

                trackingData.deleteDeletedKey(uid,true);
                trackingDataBackup.deleteDeletedKey(uid);

                Log.debug("Creating new item");
                pitem = helper.createItem();
                helper.fillItem(pitem,new String(item.getContent()));

                // Copy all the fields that are not synchronized otherwise they
                // get lost
                try {
                    helper.copyUnsupportedFields(origItem, pitem);
                } catch (Exception e) {
                    Log.error("Error while preserving unsupported fields: " + e.toString());
                    // We ignore this error and continue the synchronization
                }

                Log.trace("Before commiting PIMItem.");
                try {
                    helper.saveUpdatedPIMItem(pitem);
                } catch (PIMException e) {
                    String error = helper.handlePIMException(e);
                    logError("Unable to save updated item in source [" + getName() + "]: " + error);
                    return SyncMLStatus.GENERIC_ERROR;
                }
                // These operations must be done before changing the 
                // LUID
                trackingData.deleteNewKey(helper.getUID(pitem),true);
                trackingDataBackup.deleteNewKey(helper.getUID(pitem));
                // Preserve the old luid for this item
                setUIDForLuid(pitem, luid);

                item.setKey(luid);
                updateReceive();
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to save updated item in source [" + getName() + "]: item not found");
            return SyncMLStatus.NOT_FOUND;
        } catch(final PIMException e) {
            String error = helper.handlePIMException(e);
            logError("PIMException while updating item to SyncSource [" + getName() + "]: " + error);
            return SyncMLStatus.GENERIC_ERROR;
        } catch (Exception e) {
            // We had some sort of generic runtime exception
            logError("RuntimeException while updating item to SyncSource ["
                      + getName() + "]: " + e.toString());
            return SyncMLStatus.GENERIC_ERROR;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#deleteItem(java.lang.String)
     */
    public int deleteItem(String item) throws SyncException
    {
        Log.info("[deleteItem] " + item);

        if (app.isCancelled())  cancelSync();
        totalDeleted++;
        try
        {
            String luid = item;
            String uid = getUIDFromLuid(luid);
            Log.debug("Before trying to find and delete existing PIMItem.");
            if (helper.deleteItemFromKey(uid)!=null)
            {
                Log.trace("Purposfully deleting deleted item key: " + uid);
                trackingData.deleteDeletedKey(uid, true);
                trackingDataBackup.deleteDeletedKey(uid);

                removeLuidUIDMap(luid);
                
                updateReceive();
                return SyncMLStatus.SUCCESS;
            }
            logError("Unable to delete item in source [" + getName() + "]: no error");
            return SyncMLStatus.NOT_FOUND;
        }
        catch(final Exception e)
        {
            logError("Exception while deleting item from SyncSource ["+getName()+"]");
            return SyncMLStatus.GENERIC_ERROR;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see com.funambol.syncml.spds.SyncSource#getItemContent(com.funambol.syncml.spds.SyncItem)
     */
    protected SyncItem getItemContent(SyncItem item) throws SyncException
    {
        Log.info("[getItemContent] " + item.getKey());
        if (app.isCancelled())  cancelSync();
        try
        {
            String luid = item.getKey();
            String uid = getUIDFromLuid(luid);
            Log.debug("luid=" + luid + " uid = " + uid);
            PIMItem pitem = helper.findItem(uid);
            if (pitem!=null)
            {
                item.setContent(helper.itemToString(pitem).getBytes());
                return item;
            }
        }
        catch(final PIMException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,
                                    "Exception while retrieving content of item from SyncSource ["
                                    +getName()+"]");
        }
        return null;
    }

    private boolean initLuidUIDMap() throws SyncException
    {
        try
        {
            Log.info("[initLuidUIDMap]");
            final PersistentObject pobj = PersistentStore.getPersistentObject(luidUidPersistenceStorageKey);
            if (pobj.getContents()==null)
            {
                uidMap = new LuidUIDMap();
            }
            else
            {
                uidMap = (LuidUIDMap)pobj.getContents();
            }
            return true;
        }
        catch(Exception e)
        {
            Log.error("Cannot load LuidUID Map: " + e.toString());
            throw new SyncException(SyncException.CLIENT_ERROR, "Cannot create LuidUID map");
        }
    }


    private boolean persistLuidUID() throws SyncException
    {
        Log.info("[persistLuidUID]");
        try
        {
            final PersistentObject pobj = PersistentStore.getPersistentObject(luidUidPersistenceStorageKey);
            pobj.setContents(uidMap);
            pobj.commit();
            return true;
        }
        catch(ControlledAccessException e)
        {
            throw new SyncException(SyncException.CLIENT_ERROR,
                                    "The application could not persist LuidUID map");
        }
    }
 
    /**
     * This class holds the luid-uid mapping and can be persisted to the BB
     * storage.
     */
    public static final class LuidUIDMap implements Persistable {
        private Hashtable uidMap;

        public LuidUIDMap() {
            uidMap = new Hashtable();
        }

        public String put(String luid, String uid) {
            return (String)this.uidMap.put(luid, uid);
        }

        public String get(String luid) {
            return (String)this.uidMap.get(luid);
        }

        public Enumeration keys() {
            return this.uidMap.keys();
        }

        public void remove(String luid) {
            this.uidMap.remove(luid);
        }
    }

    private void setUIDForLuid(PIMItem item, String luid) throws PIMException, SyncException {
        Log.info("[setUIDForLuid] " + luid + " --> " + helper.getUID(item));
        uidMap.put(luid, helper.getUID(item));
        persistLuidUID();
    }

    private String getUIDFromLuid(String luid) throws PIMException, SyncException {
        String uid = (String)uidMap.get(luid);
        if (uid == null) {
            uid = luid;
        }
        Log.debug("[getUIDFromLuid] luid=" + luid + " uid="+uid);
        return uid;
    }

    protected String getLuidFromUID(String uid) {
        // We must do a reverse search in the map
        Enumeration keys = uidMap.keys();
        String luid = uid;
        while(keys.hasMoreElements()) {
            String key = (String)keys.nextElement();
            String u   = (String)uidMap.get(key);
            if (uid.equals(u)) {
                luid = key;
                break;
            }
        }
        Log.debug("[getUIDFromLuid] uid=" + uid + " luid="+luid);
        return luid;
    }

    private void removeLuidUIDMap(String luid) {
        Log.debug("[removeLuidUIDMap] " + luid);
        uidMap.remove(luid);
        persistLuidUID();
    }
}
